/* $Header: /CVSROOT/scylla-charybdis/charybdis.c,v 1.3 2003/11/06 06:57:49 tino Exp $
 *
 * $Log: charybdis.c,v $
 * Revision 1.3  2003/11/06 06:57:49  tino
 * Customizeable timeout added (300 was not enough for me now ..)
 *
 * Revision 1.2  2003/11/02 22:21:37  tino
 * CVS keywords added
 */
#define	HERE_CHARYBDIS

#define	CHARYBDIS_TIMEOUT	999	/* Increased timeout	*/
int	charybdis_timeout;

#include "sc-include.h"
#include "sc-lib.h"
#include "sc-ssl.h"
#include "sc-slist.h"

static int	throttle;
static int	maxblock	= BUFSIZ;
static int	alarm_seconds;
static int	alarm_count;
static int	alarm_timeout;
static int	alarm_signal;

static unsigned long	copycount, copycount_in, copycount_out;
static int	copy_filestart;

static void	alarm_ticker(void);

#define	USE_SIGACTION

static void
alarmhandler(int sig)
{
#ifndef USE_SIGACTION
  if (signal(SIGALRM, alarmhandler))
    ex("signal");
#endif
  alarm(alarm_seconds);
  if (++alarm_signal>alarm_count)
    ex("alarm");
}

static void
alarm_init(int s, int c)
{
#ifdef USE_SIGACTION
  struct sigaction sa;

  memset(&sa, 0, sizeof sa);
  sa.sa_handler	= alarmhandler;

  if (sigaction(SIGALRM, &sa, NULL))
    ex("sigaction");
#else
  signal(SIGALRM, alarmhandler);
#endif

  alarm_seconds	= s;
  alarm_count	= c;
  alarm_timeout	= 0;
  alarm_signal	= 0;

  alarm(s);
}

static void
alarm_process(void)
{
  if (alarm_signal)
    {
      if (++alarm_timeout>alarm_count)
	ex("timeout");

      alarm_ticker();

      alarm_signal	= 0;
    }
}

static void
alarm_reset(void)
{
  alarm_timeout	= 0;
}

static void
writen(const char *buf, size_t len, int fo)
{
  static long		delta;
  static struct timeval	last;
  int			pos, put;

  alarm_reset();
  for (pos=0; pos<len; pos+=put)
    {
      size_t	max;

      max	= len-pos;
      if (max>maxblock)
	max	= maxblock;

      put	= tino_ssl_write(fo, buf+pos, max);
      alarm_process();

      if (put>0)
	{
	  if (fo==sock)
	    copycount_out	+= put;
	  if (put>max)
	    ex("oops");
	  alarm_reset();
	  if (throttle)
	    {
	      struct timeval	now;
	      long		hold, elapsed;

	      /* This code is self synchronizing
	       * to any type of problem.
	       */
	      gettimeofday(&now, NULL);
	      elapsed	=  (now.tv_sec-last.tv_sec)*1000+(now.tv_usec-last.tv_usec)/1000;
	      last	=  now;
	      if (elapsed<0)
		elapsed	= 0;
	      if (elapsed>1000000)
		elapsed	= 1000000;
	      if (copy_filestart)
		{
		  copy_filestart	= 0;
		  elapsed		= 0;
		}
	      delta	+= put;
	      delta	-= elapsed*throttle;
	      if (delta<0)
		{
		  if (delta<-maxblock)
		    delta	=  -maxblock;
		}
	      else
		{
	          hold	=  delta*1000/throttle;
	          if (hold>10000)
		    {
#if 0
		      fprintf(stderr, " t=%ld d=%ld n=%ld\r", hold/1000, delta, elapsed);
#endif
		      fflush(stderr);
		      usleep(hold);
		    }
                }
	    }
	  continue;
	}

      if (!put)
	ex("premature EOF on output");

      put	= 0;
      if (errno!=EINTR && errno!=EAGAIN)
	ex("output");
    }
}

static int
readn(int fd, void *buf, size_t max)
{
  alarm_reset();
  for (;;)
    {
      int	got;

      got	= tino_ssl_read(fd, buf, max);
      alarm_process();

      if (got>=0)
	{
	  alarm_reset();
	  if (fd==sock)
	    copycount_in	+= got;
	  return got;
	}

      if (errno!=EINTR && errno!=EAGAIN)
	ex("input");
    }
}

#ifdef NO_ZLIB
static int
docopy(int fi, int fo, int ignore)
{
  char		*buf;
  int		got;
  size_t	bs;

  copy_filestart	= 1;
  bs	= maxblock;
  if (bs<BUFSIZ)
    bs	= BUFSIZ;
  buf	= re_alloc(NULL, bs);
  while ((got=readn(fi, buf, bs))!=0)
    {
      writen(buf, got, fo);
      copycount		+= got;
    }
  free(buf);
  return 0;
}
#else
#include "zlib.h"
static void *
dozalloc(void *ign, unsigned item, unsigned size)
{
  return re_alloc(NULL, (long)item*(long)size);
}
static void
dozfree(void *ign, void *p)
{
  free(p);
}
static int
docopy(int fi, int fo, int doinflate)
{
  char		*bi, *bo;
  size_t	bs;
  z_stream	str;
  int		init, mode, ret;

  copy_filestart	= 1;
  bs	= maxblock;
  if (bs<BUFSIZ)
    bs	= BUFSIZ;
  bi	= re_alloc(NULL, bs);
  bo	= re_alloc(NULL, bs);

  str.zalloc	= dozalloc;
  str.zfree	= dozfree;
  str.opaque	= 0;

  str.avail_in	= 0;
  str.next_in	= bi;

  str.avail_out	= bs;
  str.next_out	= bo;

  init		= 1;
  mode		= Z_NO_FLUSH;

  do
    {
      if (!str.avail_in && mode==Z_NO_FLUSH)
	{
	  if ((str.avail_in=readn(fi, str.next_in=bi, bs))==0)
	    mode	= Z_FINISH;
	  copycount	+= str.avail_in;
	}

      if (init)
	{
	  init	= 0;

	  if (doinflate)
	    {
	      if (inflateInit(&str)!=Z_OK)
		ex("inflate");
	    }
	  else if (deflateInit(&str, 9)!=Z_OK)
	    ex("deflate");
	}

      ret	= (doinflate ? inflate : deflate)(&str, mode);
#if 0
      fprintf(stderr, "\nmode=%d ret=%d\n", mode, ret);
#endif

      if (!str.avail_out || mode!=Z_NO_FLUSH || ret==Z_STREAM_END)
	{
	  size_t	put;

	  put	= bs-str.avail_out;
	  if (put)
	    writen(bo, put, fo);
	  else if (mode!=Z_NO_FLUSH && ret==Z_BUF_ERROR && !str.avail_in)
	    ex("starved");
#if 0
	  fprintf(stderr, "put=%d\n", put);
#endif
	  str.avail_out	= bs;
	  str.next_out	= bo;
	}
    }
  while (ret==Z_OK || ret==Z_BUF_ERROR);

  /* 2002-07-14: Memory leak resolved	*/
  if (!init && (doinflate ? inflateEnd : deflateEnd)(&str)!=Z_OK)
     ex("zlib err");

  free(bi);
  free(bo);

  return ret!=Z_STREAM_END;
}
#endif

/* That's not optimal, however we do not send big things this way.
 */
static void
swrite(const char *s)
{
  size_t	len;
  char		*tmp;

  len	= strlen(s);
  tmp	= alloc(len+1);
  memcpy(tmp, s, len);
  tmp[len]='\n';
  writen(tmp, len+1, sock);
  free(tmp);
}

static char *
sread(void)
{
  char		buf[1024];
  char		*out	= 0;
  size_t	len	= 0;
  int		got;

  while ((got=readn(sock, buf, sizeof buf))>0)
    {
      char	*tmp;

      out	= re_alloc(out, len+got+1);
      memcpy(out+len, buf, got);
      out[len+got]	= 0;
      if ((tmp=strchr(out+len, '\n'))!=0)
	{
	  *tmp	= 0;
	  return out;
	}
      len	+= got;
    }
  out		= re_alloc(out, len+1);
  out[len]	= 0;
  return out;
}

static int
sread_match(const char *s, int n)
{
  char	*t;
  int	res;

  if (!n)
    n	= strlen(s)+1;
  t	= sread();
  res	= strncmp(t, s, n);
  free(t);
  return res;
}

#include "sc-md5.h"

#include <netinet/tcp.h>
#include <time.h>
#include <gdbm.h>

static unsigned long	totalcount, overflowcount;
static char		workmode;
static int		returnval;
static int		filecount, filecopy;

static GDBM_FILE	gdbm;

#define	INTERPOLATION 30

static void
alarm_ticker(void)
{
  static int	speedometer[INTERPOLATION], speed2[INTERPOLATION];
  static time_t	timing[INTERPOLATION];
  static unsigned long	lastcount, lastio;
  static unsigned long	gigabytes;
  static long	delta, delta2;
  static int	pos;
  time_t	now;
  long		seconds;
  double	bytes, bytes2, percent;
  int		e;

  e	= errno;
  pos			= (pos+1)%INTERPOLATION;

  time(&now);

  if (overflowcount)
    {
      lastcount		= totalcount-overflowcount;
      overflowcount	= 0;
    }
  if (copycount_out>=(1ul<<30))
    {
      gigabytes++;
      copycount_out	-= 1ul<<30;
      lastio		-= 1ul<<30;
    }
  delta2		-= speed2[pos];
  delta2		+=
    speed2[pos]		= copycount_out-lastio;
  lastio		= copycount_out;

  delta			-= speedometer[pos];
  delta			+= 
    speedometer[pos]	= copycount-lastcount;
  lastcount		= copycount;

  seconds		= now-(timing[pos] ? timing[pos] : timing[1]);
  timing[pos]		= now;
  if (seconds<=0)
    seconds		= 1;

  bytes			= (double)delta/seconds/1024.;
  bytes2		= (double)delta2/seconds/1024.;
  percent		= -1;
  if (totalcount)
    percent		= copycount*100./totalcount;

  printf("\r%c %7ld %6.2f%% %8.2f %8.2f %2d %d %d %lu.%05lx",
	 workmode, (totalcount-copycount)>>10,
	 percent, bytes, bytes2, alarm_timeout,
	 filecount, filecopy,
	 gigabytes, copycount_out>>10);
  fflush(stdout);
  errno	= e;
}

static void
getaddr(struct sockaddr_in *sin, const char *host, int port)
{
  memset(sin, 0, sizeof *sin);
  sin->sin_family	= AF_INET;
  sin->sin_addr.s_addr	= htonl(INADDR_ANY);
  sin->sin_port		= htons(port);
  if (host && !inet_aton(host, &sin->sin_addr))
    {
      struct hostent	*he;

      if ((he=gethostbyname(host))==0)
	ex(host);
      if (he->h_addrtype!=AF_INET || he->h_length!=sizeof sin->sin_addr)
	ex("unsupported host address");
      memcpy(&sin->sin_addr, he->h_addr, sizeof sin->sin_addr);
    }
}

static void
sockopen(const char *host, int port)
{
  struct sockaddr_in	sin;
  int			on;
  struct linger		linger;
  unsigned int		mss;

  sock	= socket( AF_INET, SOCK_STREAM, 0 );
  if (sock<0)
    ex("socket");

  on = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on))
    ex("setsockopt reuse");
  linger.l_onoff	= 1;
  linger.l_linger	= 65535;
  if (setsockopt(sock, SOL_SOCKET, SO_LINGER, &linger, sizeof linger))
    ex("setsockopt linger");
  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &on, sizeof on))
    ex("setsockopt sndbuf");
  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &on, sizeof on))
    ex("setsockopt rcvbuf");

  /* 2002-05-09 set the MTU
   * The kernel takes a lower value if it is not supported ;)
   * We add some bytes for the tcp packet overhead.
   */
  if (maxblock<1500)
    {
      mss = maxblock+48;
      if(setsockopt(sock, IPPROTO_TCP, TCP_MAXSEG, &mss, sizeof(mss)))
        ex("setsockopt maxseg");
    }

  getaddr(&sin, host, port);

  if (bind(sock, (struct sockaddr *)&sin, sizeof sin))
    ex("bind");

#if 0
  if (fcntl(sock, F_SETFL, O_NDELAY))
    ex("O_NDELAY");
#endif
}

static void
do_connect(const char *host, int port)
{
  struct sockaddr_in	sin;

  sockopen(NULL, 0);

  getaddr(&sin, host, port);
  if (connect(sock, (struct sockaddr *)&sin, sizeof sin))
    ex("connect");
}

/* We don't want to connect to scylla
 * except we have to do something.
 * Therefor this delayed connect function.
 * Additionally this prohibits a timeout
 * until find feeds us the files ;)
 */
void
delayed_connect(const char *host, int port, const char *pass)
{
  static char	*sav_host, *sav_pass;
  static int	sav_port;

  DP(("delayed_connect(%s,%d,%s)", host, port, pass));
  if (host)
    {
      sav_host	= str_alloc(host);
      sav_port	= port;
      sav_pass	= str_alloc(pass);
      sock	= -1;
      return;
    }
  if (sock>=0)
    return;

  workmode	= '0';
  alarm_init(1 /* see below */, charybdis_timeout);

  /* Hack: Set a connect timeout to 5 minutes
   */
  alarm(300);
  do_connect(sav_host, sav_port);

#ifndef NO_SSL
  alarm_process();
  alarm(300);
  tino_ssl_client();
#endif

  /* Reset timeout to alarm processing
   */
  alarm_reset();
  alarm_process();
  alarm(1 /* see above */);

  workmode	= '1';

  /* send password */
  swrite(sav_pass);

  if (sread_match("scylla", 6))
    ex("where is my dear scylla?");
  /* leave version control for future
   * we are using version 1.0 protocol
   * (plain password)
   */
}

static int
dofile(const char *file)
{
  int		fd;
  char		*t;
  unsigned long	len;
  struct stat	st;

  DP(("dofile(%s)", file));

  filecount++;
  alarm_reset();
  alarm_process();

  if ((fd=open(file, O_RDONLY))<0)
    {
      perror(file);
      DP(("dofile() !file"));
      returnval	|= 1;
      return -1;
    }
  if (fstat(fd, &st) ||
      !S_ISREG(st.st_mode))
    {
      perror(file);
      DP(("dofile() !fstat"));
      close(fd);
      returnval	|= 1;
      return -1;
    }

  /* Say there is more on streaming mode
   * On the first file delayed_connect is done,
   * thus sock<0 here
   */
  if (sock>=0)
    swrite("m");

  /* Now that we have something to do: connect
   */
  delayed_connect(NULL, 0, NULL);

  workmode	= '2';
  swrite(*file=='/' ? file+1 : file);

  t	= sread();
  if (!isdigit(*t))
    {
      len	= 0;
      swrite("c");
      workmode	= 'c';
    }
  else
    {
      char	*m;

      sscanf(t, "%lu", &len);
      m	= md5sum(fd, len);
      if (!m)
	ex("sum");

      if (!strcmp(m, t))
	{
	  if (st.st_size==len)
	    {
	      free(m);
	      free(t);
	      close(fd);
	      swrite("o");
	      /* unlink(argv[4]); */
	      return 0;
	    }
	  swrite("a");
	  workmode	= 'a';
	}
      else
	{
	  len	= 0;
	  swrite("w");
	  workmode	= 'w';
	}
      free(m);
    }
  free(t);

  filecopy++;

  if (lseek(fd, len, SEEK_SET)!=len)
    ex("lseek");
  overflowcount	+= copycount;
  copycount	= 0;
  totalcount	= st.st_size-len;

  if (sread_match("OK", 0))
    ex("protocol failure");

  if (docopy(fd, sock, 0))
    ex("copy error");
  if (close(fd))
    ex("close error");

  returnval	|= 2;
  workmode	= 'e';
  /* Wait for transfer flush */

  if (sread_match("NEXT", 0))
    {
      fprintf(stderr, "protocol mismatch: missing next\n");
      exit(returnval);
    }
  return 1;
}

/* TODO in 2.3:
 * We are not allowed to buffer, as if something breaks,
 * we must leave the other files on stdin.
 * Therefor we must loop over bytes.
 * Fix this with S&C version 3 (or 4?).
 */
void
stdinfileloop(void)
{
  char	buf[BUFSIZ];

  000;
  DP(("stdinfileloop"));
  while (fgets(buf, sizeof buf-1, stdin))
    {
      size_t	len;

      len	= strlen(buf);
      if (len>=sizeof buf-1)
	ex("filename too long");
      if (len && buf[len-1]=='\n')
	buf[--len]	= 0; 
      if (len<1)
	continue;

      dofile(buf);
    }
}

static void
gdbmfatalfn(const char *s)
{
  ex(s);
}

static void
close_dbm(void)
{
  if (gdbm)
    gdbm_close(gdbm);
  gdbm	= 0;
}

static GDBM_FILE
open_dbm(char *name)
{
  workmode	= 'G';
  alarm_reset();
  close_dbm();
  for (;;)
    {
      gdbm		= gdbm_open(name, BUFSIZ, GDBM_READER|GDBM_WRITER|GDBM_SYNC, 0, gdbmfatalfn);
      alarm_process();
      if (gdbm)
	return gdbm;
      usleep(10000ul);
    }
  workmode	= 'g';
}

static int
iswork(datum k)
{
  datum	d;
  int	tmp;

  d	= gdbm_fetch(gdbm, k);
  if (!d.dptr)
    return 0;
  tmp	= strncmp(d.dptr, "charybdis", d.dsize);
  free(d.dptr);
  if (d.dsize<9 || tmp)
    return 0;
  return 1;
}

static void
gdbmfileloop(char *name)
{
  char	buf[BUFSIZ];
  int	round;
  SLIST	list;
  int	n;

  alarm_init(1, charybdis_timeout);
  round		= -1;

  list	= slist_new();
  for (;;)
    {
      datum		k, d;
      const char	*s;
      void		*old;
      int		r;

      /* find some work to do
       * put it into slist
       */
      open_dbm(name);
      n	= 0;
      r	= 0;
      for (k=gdbm_firstkey(gdbm); k.dptr; old=k.dptr, k=gdbm_nextkey(gdbm, k), free(old))
	{
#if 0
	  if (~((++r)&0x7)==0)
	    {
	      fprintf(stderr, "\r%d", r);
	      fflush(stderr);
            }
#endif
          if (!iswork(k) ||
	      k.dsize>=sizeof buf)
            continue;

          memcpy(buf, k.dptr, k.dsize);
          buf[k.dsize]	= 0;

	  slist_add(list, buf);
	  if (++n>=1000)
	    {
	      free(k.dptr);
	      break;
	    }
	}

      fprintf(stderr, "\r\ngot worklist %d\n", n);
      if (!n)
	break;

      while ((s=slist_get(list))!=0)
	{
	  int	tmp;

	  k.dptr	= s;
	  k.dsize	= strlen(s);
	  if (!iswork(k))
	    continue;

	  d.dptr	= "work";
	  d.dsize	= 4;
	  if (gdbm_store(gdbm, k, d, GDBM_REPLACE))
	    continue;

          close_dbm();
          tmp	= dofile(s);
          open_dbm(name);

          d.dptr	= (tmp<0 ? "ko" : tmp>0 ? "do" : "ok");
          d.dsize	= 2;
          if (gdbm_store(gdbm, k, d, GDBM_REPLACE))
	    ex("store");
        }
    }
}

int
main(int argc, char **argv)
{
  DP(("main(%d,%p)", argc, argv));
  if (argc<5)
    {
      printf("Usage: %s host port password filename [throttle [blocksize [timeout]]]\n"
	     "\tCharybdis assumes that Scylla is at host:port and authenticates with password.\n"
	     "\tWhen the file is uploaded to Scylla it returns 2 or\n"
	     "\tif Scylla already has the file completely it returns 0.\n"
	     "\tThe bandwidth can be controlled with trottle@blocksize (throttle is in KB/s)\n"
	     "\tTimeout defaults to 999 seconds\n"
	     "\tIf file is - then the file list is read from stdin.  0 means all files are ok\n"
	     "\tIf file is gdbm:file then files are keys in a gdbm file with value 'charybdis'.\n"
	     "\tValue will be changed to 'do', 'ok' or 'ko' accordingly (in progess is 'work').\n"
             ,
	     argv[0]);
      return 1;
    }

  returnval	= 0;
  if (argc>5)
    {
      throttle	= atoi(argv[5]);
      maxblock	= 512;
      if (argc>6)
	maxblock	= atoi(argv[6]);
    }
  if (maxblock<100)
    maxblock	= BUFSIZ;
  else if (maxblock<256)
    maxblock	= 256;

  charybdis_timeout	= CHARYBDIS_TIMEOUT;
  if (argc>7)
    charybdis_timeout	= atoi(argv[7]);
  if (charybdis_timeout<=0)
    charybdis_timeout	= CHARYBDIS_TIMEOUT;

  delayed_connect(argv[1], atoi(argv[2]), argv[3]);

  if (!strcmp(argv[4], "-"))
    stdinfileloop();
  else if (!strncmp(argv[4], "gdbm:", 5))
    gdbmfileloop(argv[4]+5);
  else
    dofile(argv[4]);

  if (sock>=0)
    {
      swrite("e");

#ifndef NO_SSL
      tino_ssl_close();
#endif
      close(sock);
    }
  return returnval;
}
